#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef USE_LIBEDIT
#include <editline/readline.h>
#endif
#ifdef USE_EDITLINE
#include <editline.h>
#endif
#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif
#ifdef USE_NLLINE
#include <nlline.h>
#endif
#if defined(USE_GETLINE) || defined(USE_FGETS)
#include "readline.h"
#define readline(prompt) nlsh_readline(prompt)
#define add_history(line) nlsh_add_history(line)
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "args.h"
#include "val.h"

static int sigint_handled = 0;
static int sigchld_handled = 0;

static void signal_reset(void)
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_DFL;
  sigaction(SIGINT, &sa, NULL);

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_DFL;
  sigaction(SIGCHLD, &sa, NULL);
}

static int exec_command(char *argv[], char *envp[])
{
  int n, i, flags, fd, out, save[3];
  char **args = NULL;
  char *outstr = NULL;
  enum {
    EXEC_ERR_UNKNOWN = 0,
    EXEC_ERR_NOFILE,
    EXEC_ERR_FILEERR,
  } e = EXEC_ERR_UNKNOWN;

  for (i = 0; i < 3; i++)
    save[i] = -1;

  while (1) {
    n = search_argv(argv, "<");
    if (n < 0)
      break;
    args = extract_argv(argv, n, 2);
    if (get_argc(args) < 2) {
      e = EXEC_ERR_NOFILE;
      goto err;
    }
    fd = open(args[1], O_RDONLY);
    if (fd < 0) {
      e = EXEC_ERR_FILEERR;
      goto err;
    }
    if (save[0] < 0)
      save[0] = dup(0);
    close(0);
    dup2(fd, 0);
    close(fd);
    free_argv(args);
    args = NULL;
  }

  while (1) {
    n = search_argv(argv, ">");
    if (n < 0) n = search_argv(argv, "2>");
    if (n < 0) n = search_argv(argv, ">>");
    if (n < 0) n = search_argv(argv, "2>>");
    if (n < 0)
      break;
    args = extract_argv(argv, n, 2);
    if (get_argc(args) < 2) {
      e = EXEC_ERR_NOFILE;
      goto err;
    }
    if (!strcmp(args[0], ">>") || !strcmp(args[0], "2>>")) {
      flags = O_WRONLY | O_CREAT | O_APPEND;
    } else {
      flags = O_WRONLY | O_CREAT | O_TRUNC;
    }
    if (!strcmp(args[0], "2>") || !strcmp(args[0], "2>>")) {
      out = 2;
    } else {
      out = 1;
    }

    fd = open(args[1], flags, 0644);
    if (fd < 0) {
      e = EXEC_ERR_FILEERR;
      goto err;
    }
    if (save[out] < 0)
      save[out] = dup(out);
    close(out);
    dup2(fd, out);
    close(fd);
    free_argv(args);
    args = NULL;
  }

  do {
    n = search_argv(argv, "1>&2");
    if (n < 0) n = search_argv(argv, "2>&1");
    if (n < 0)
      break;
    args = extract_argv(argv, n, 1);
    if (!strcmp(args[0], "1>&2")) {
      if (save[1] < 0)
	save[1] = dup(1);
      close(1);
      dup2(2, 1);
    } else {
      if (save[2] < 0)
	save[2] = dup(2);
      close(2);
      dup2(1, 2);
    }
    free_argv(args);
    args = NULL;
  } while (n >= 0);

  if (get_argc(argv) == 0) {
    goto err;
  }

  signal_reset();

  execve(argv[0], argv, envp);

err:
  if (args)
    free_argv(args);

  for (i = 0; i < 3; i++) {
    if (save[i] >= 0) {
      dup2(save[i], i);
      close(save[i]);
    }
  }

  switch (e) {
  case EXEC_ERR_NOFILE:  outstr = "No file selected.\n"; break;
  case EXEC_ERR_FILEERR: outstr = "File not found.\n"; break;
  case EXEC_ERR_UNKNOWN:
  default:
    outstr = "Unknown error.\n";
    break;
  }
  if (outstr)
    write(2, outstr, strlen(outstr));

  return -1;
}

static int exec_line(char *line, char *envp[])
{
  int argc, n, status, r;
  char **argv, **exec;
  char *outstr;
  pid_t fpid = -1, lpid = -1, pid;
  int fildes[2], outfd = -1, pipe_in, bg, fg_cnt = 0, bg_cnt = 0;

  argv = make_argv(line);
  argc = get_argc(argv);

  r = argc;

  while (argc > 0) {
    pipe_in = 0;
    n = rsearch_argv(argv, "|");
    if (n < 0) {
      exec = extract_argv(argv, 0, argc);
    } else {
      exec = extract_argv(argv, n, -1);
      delete_argv(exec, 0, 1);
    }

    argc = get_argc(argv);

    bg = rsearch_argv(exec, "&");
    if (bg >= 0)
      delete_argv(exec, bg, 1);

    if (get_argc(exec) == 0) {
      continue;
    }

    if (!strcmp(exec[0], "cd") && (get_argc(exec) > 1)) {
      if (chdir(exec[1]) < 0) {
	write(2, exec[1], strlen(exec[1]));
	write(2, ": No such file or directory.\n", 29);
      }
      continue;
    }

    if (argc > 0) {
      pipe2(fildes, 0);
      pipe_in = 1;
    }

    if (bg < 0)
      fg_cnt++;
    else
      bg_cnt++;

    pid = fork();
    if (pid == 0) {
      if (outfd >= 0) {
	close(1);
	dup2(outfd, 1);
	close(outfd);
      }
      if (pipe_in) {
	close(fildes[1]);
	close(0);
	dup2(fildes[0], 0);
	close(fildes[0]);
      }
      if (exec_command(exec, envp) < 0)
	_exit(1);
      _exit(0);
    }

    if (fpid < 0)
      fpid = pid;
    lpid = pid;

    outfd = -1;
    if (pipe_in) {
      close(fildes[0]);
      outfd = fildes[1];
    }

    free_argv(exec);
  }

  while (fg_cnt > 0) {
    sigint_handled = 0;
    pid = wait4(-1, &status, 0, NULL);

    if (pid < 0) {
      if (errno == ECHILD) {
	break;
      }
      if (errno == EINTR) {
	if (sigint_handled) {
	  if (lpid >= 0)
	    kill(lpid, SIGINT);
	}
      }
      continue;
    }

    outstr = "Unknown.\n";
    if (WIFSTOPPED(status)) {
      outstr = "Stopped.\n";
    } else if (WIFEXITED(status)) {
      outstr = "Exited.\n";
      fg_cnt--;
    } else if (WIFCONTINUED(status)) {
      outstr = "Cont.\n";
    } else if (WIFSIGNALED(status)) {
      outstr = "Received signal.\n";
      switch (WTERMSIG(status)) {
      case SIGINT: fg_cnt--; break;
      default:
	break;
      }
    }
    write(2, outstr, strlen(outstr));
  }

  free_argv(argv);

  return r;
}

void sigint_handler(int arg)
{
  sigint_handled++;
  return;
}

void sigchld_handler(int arg)
{
  sigchld_handled++;
  return;
}

static void signal_init(void)
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigint_handler;
  sigaction(SIGINT, &sa, NULL);

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigchld_handler;
  sigaction(SIGCHLD, &sa, NULL);
}

static void init(void)
{
  signal_init();
}

static char *make_prompt(void)
{
  char *p, *user, *home, *host, *u;
  char cwd[256], hostname[256];
  static char *prompt = NULL;

  user = val_get("USER");
  home = val_get("HOME");
  host = val_get("HOST");
  if (getcwd(cwd, sizeof(cwd)) == NULL)
    cwd[0] = 0;

  u = user ? user : "unknown";

  if ((host != NULL) && (strlen(host) < sizeof(hostname))) {
    strcpy(hostname, host);
    p = strchr(hostname, '.');
    if (p)
      *p = '\0';
  }

  if ((home != NULL) && (strlen(home) > 1) &&
      !memcmp(cwd, home, strlen(home))) {
    memmove(cwd + 1, cwd + strlen(home), strlen(cwd) - strlen(home) + 1);
    cwd[0] = '~';
  }

  prompt = realloc(prompt, 32 + strlen(u) + strlen(hostname) + strlen(cwd));
  if (prompt != NULL) {
    sprintf(prompt, "%s@%s:%s>%% ", u, hostname, cwd);
  }

  return prompt ? prompt : "> ";
}

int main(int argc, char *argv[], char *envp[])
{
  char *line, *outstr, *prompt;
  int i, finished = 0, status;
  pid_t pid;

#if 0
  for (i = 0; envp[i]; i++)
    printf("envp[%d]: %s\n", i, envp[i]);
  for (i = 0; i < argc; i++)
    printf("argv[%d]: %s\n", i, argv[i]);
#endif

  init();

  for (i = 0; envp[i]; i++) {
    val_add_from_string(envp[i]);
  }

  while (!finished) {
    prompt = make_prompt();

    line = readline(prompt);
    if (line == NULL) {
      write(1, "\n", 1);
      while (1) {
	pid = wait4(-1, &status, WNOHANG, NULL);
	if (pid <= 0)
	  break;
	outstr = "Unknown.\n";
	if (WIFSTOPPED(status))        outstr = "Stopped.\n";
	else if (WIFEXITED(status))    outstr = "Exited.\n";
	else if (WIFCONTINUED(status)) outstr = "Cont.\n";
	else if (WIFSIGNALED(status))  outstr = "Received signal.\n";
	if (outstr)
	  write(2, outstr, strlen(outstr));
      }
      continue;
    }

    if (!strcmp(line, "exit"))
      finished = 1;
    else if (exec_line(line, envp) > 0)
      add_history(line);

    free(line);
  }

  return 0;
}
